"""Module for constructing and manipulating CSS codes."""

from __future__ import annotations

from .bb_codes import construct_bb_code
from .color_code import ColorCode, LatticeType
from .css_code import CSSCode, InvalidCSSCodeError
from .hexagonal_color_code import HexagonalColorCode
from .square_octagon_color_code import SquareOctagonColorCode

__all__ = [
    "CSSCode",
    "ColorCode",
    "HexagonalColorCode",
    "InvalidCSSCodeError",
    "LatticeType",
    "SquareOctagonColorCode",
    "construct_bb_code",
]



"""Simulation utilities for analog information decoding."""

from __future__ import annotations

import json
import locale
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
from ldpc.mod2 import rank
from scipy.special import erfc, erfcinv

from .data_utils import BpParams, calculate_error_rates, replace_inf

if TYPE_CHECKING:
    from numpy.typing import NDArray


def set_seed(value: float) -> None:
    """The appropriate way to set seeds when numba is used."""
    np.random.seed(value)  # noqa: NPY002


def alist2numpy(fname: str) -> NDArray[np.int32]:  # current original implementation is buggy
    """Converts an alist file to a numpy array."""
    alist_file: NDArray[np.str_] = np.loadtxt(fname, delimiter=",", dtype=str)
    matrix_dimensions = alist_file[0].split()
    m = int(matrix_dimensions[0])
    n = int(matrix_dimensions[1])

    mat: NDArray[np.int32] = np.zeros((m, n), dtype=np.int32)

    for i in range(m):
        columns = [item for item in alist_file[i + 4].split() if item.isdigit()]
        columns_two: NDArray[np.int32] = np.array(columns, dtype=np.int32)
        columns_two -= 1  # convert to zero indexing
        mat[i, columns_two] = 1

    return mat


# Rewrite such that call signatures of check_logical_err_h
# and check_logical_err_l are identical
def check_logical_err_h(
    check_matrix: NDArray[np.int_], original_err: NDArray[np.int_], decoded_estimate: NDArray[np.int_]
) -> bool:
    """Checks if the residual error is a logical error."""
    _, n = check_matrix.shape

    # compute residual err given original err
    residual_err: NDArray[np.int32] = np.zeros((n, 1), dtype=np.int32)
    for i in range(n):
        residual_err[i][0] = original_err[i] ^ decoded_estimate[i]

    ht = np.transpose(check_matrix)

    htr = np.append(ht, residual_err, axis=1)

    rank_ht = rank(check_matrix)  # rank A = rank A.T

    rank_htr = rank(htr)

    return (rank_ht < rank_htr) is True


# L is a numpy array, residual_err is vector s.t. dimensions match
# residual_err is a logical iff it commutes with logicals of other side
# i.e., an X residal is a logical iff it commutes with at least one Z logical and
# an Z residual is a logical iff it commutes with at least one Z logical
# Hence, L must be of same type as H and of different type than residual_err
def is_logical_err(logicals: NDArray[np.int_], residual_err: NDArray[np.int_]) -> bool:
    """Checks if the residual error is a logical error.

    :returns: True if its logical error, False otherwise (is a stabilizer).
    """
    l_check = (logicals @ residual_err) % 2
    return bool(l_check.any())  # check all zeros


# adapted from https://github.com/quantumgizmos/bp_osd/blob/a179e6e86237f4b9cc2c952103fce919da2777c8/src/bposd/css_decode_sim.py#L430
# and https://github.com/MikeVasmer/single_shot_3D_HGP/blob/master/sim_scripts/single_shot_hgp3d.cpp#L207
# channel_probs = [x,y,z], residual_err = [x,z]
def generate_err(
    nr_qubits: int,
    channel_probs: tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]],
    residual_err: list[NDArray[np.int_]],
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    """Computes error vector with X and Z part given channel probabilities and residual error.

    Assumes that residual error has two equally sized parts.
    """
    error_x = residual_err[0]
    error_z = residual_err[1]
    channel_probs_x = channel_probs[0]
    channel_probs_y = channel_probs[1]
    channel_probs_z = channel_probs[2]
    residual_err_x = residual_err[0]
    residual_err_z = residual_err[1]

    for i in range(nr_qubits):
        rand = np.random.default_rng().random()  # this returns a random float in [0,1)
        # e.g. if err channel is p = 0.3, then an error will be applied if rand < p
        if rand < channel_probs_z[i]:  # if probability for z error high enough, rand < p, apply
            # if there is a z error on the i-th bit, flip the bit but take residual error into account
            # nothing on x part - probably redundant anyways
            error_z[i] = (residual_err_z[i] + 1) % 2
        elif (  # if p = 0.3 then 0.3 <= rand < 0.6 is the same sized interval as rand < 0.3
            channel_probs_z[i] <= rand < (channel_probs_z[i] + channel_probs_x[i])
        ):
            # X error
            error_x[i] = (residual_err_x[i] + 1) % 2
        elif (  # 0.6 <= rand < 0.9
            (channel_probs_z[i] + channel_probs_x[i])
            <= rand
            < (channel_probs_x[i] + channel_probs_y[i] + channel_probs_z[i])
        ):
            # y error == both x and z error
            error_z[i] = (residual_err_z[i] + 1) % 2
            error_x[i] = (residual_err_x[i] + 1) % 2

    return error_x, error_z


def get_analog_llr(analog_syndrome: NDArray[np.float64], sigma: float) -> NDArray[np.float64]:
    """Computes analog LLRs given analog syndrome and sigma."""
    if sigma <= 0.0:
        return np.zeros_like(analog_syndrome).astype(np.float64)
    return (2 * analog_syndrome) / (sigma**2)


def get_sigma_from_syndr_er(ser: float) -> float:
    """For analog Cat syndrome noise we need to convert the syndrome error model as described in the paper.

    :return: sigma.
    """
    if ser == 0.0:
        return 0.0
    return float(1 / np.sqrt(2) / (erfcinv(2 * ser)))  # see Eq. cref{eq:perr-to-sigma} in our paper


def get_error_rate_from_sigma(sigma: float) -> float:
    """For analog Cat syndrome noise we need to convert the syndrome error model as described in the paper.

    :return: sigma.
    """
    if sigma == 0.0:
        return 0.0
    return float(0.5 * erfc(1 / np.sqrt(2 * sigma**2)))  # see Eq. cref{eq:perr-to-sigma} in our paper


def get_virtual_check_init_vals(noisy_syndr: NDArray[np.float64], sigma: float) -> NDArray[np.float64]:
    """Computes a vector of values v_i from the noisy syndrome bits y_i s.t.

    BP initializes the LLRs l_i of the analog nodes with the
    analog info values (see paper section). v_i := 1/(e^{y_i}+1).
    """
    if sigma <= 0.0:
        return np.zeros_like(noisy_syndr).astype(np.float64)
    llrs = get_analog_llr(noisy_syndr, sigma)
    return np.array(1 / (np.exp(np.abs(llrs)) + 1))


def generate_syndr_err(channel_probs: NDArray[np.float64]) -> NDArray[np.int32]:
    """Generates a random error vector given the error channel probabilities."""
    error: NDArray[np.int32] = np.zeros_like(channel_probs, dtype=np.int32)

    for i, p in np.ndenumerate(channel_probs):
        rand = np.random.default_rng().random()

        if rand < p:
            error[i] = 1

    return error


def get_noisy_analog_syndrome(perfect_syndr: NDArray[np.int_], sigma: float) -> NDArray[np.float64]:
    """Generate noisy analog syndrome vector given the perfect syndrome and standard deviation sigma (~ noise strength).

    Assumes perfect_syndr has entries in {0,1}.
    """
    # compute signed syndrome: 1 = check satisfied, -1 = check violated. float needed for Gaussian sampling call
    sgns: NDArray[np.float64] = np.where(
        perfect_syndr == 0.0,
        np.ones_like(perfect_syndr),
        np.full_like(perfect_syndr, -1.0),
    ).astype(float)

    # sample from Gaussian with zero mean and sigma std. dev: ~N(0, sigma_sq)
    return np.array(np.random.default_rng().normal(loc=sgns, scale=sigma, size=perfect_syndr.shape)).astype(np.float64)


def error_channel_setup(
    error_rate: float, xyz_error_bias: NDArray[np.float64], nr_qubits: int
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Set up an error_channel given the physical error rate, bias, and number of bits."""
    xyz_error_bias = np.array(xyz_error_bias)
    if xyz_error_bias[0] == np.inf:
        px = error_rate
        py = 0.0
        pz = 0.0
    elif xyz_error_bias[1] == np.inf:
        px = 0.0
        py = error_rate
        pz = 0.0
    elif xyz_error_bias[2] == np.inf:
        px = 0.0
        py = 0.0
        pz = error_rate
    else:
        px, py, pz = (
            error_rate * xyz_error_bias / np.sum(xyz_error_bias)
        )  # Oscar only considers X or Z errors. For reproducibility remove normalization

    channel_probs_x = np.ones(nr_qubits) * px
    channel_probs_z = np.ones(nr_qubits) * pz
    channel_probs_y = np.ones(nr_qubits) * py

    return channel_probs_x, channel_probs_y, channel_probs_z


def build_single_stage_pcm(pcm: NDArray[np.int_], meta: NDArray[np.int_]) -> NDArray[np.int_]:
    """Build the single statge parity check matrix."""
    id_r = np.identity(meta.shape[1])
    zeros = np.zeros((meta.shape[0], pcm.shape[1]))
    return np.block([[pcm, id_r], [zeros, meta]])


def get_signed_from_binary(binary_syndrome: NDArray[np.int_]) -> NDArray[np.int_]:
    """Maps the binary vector with {0,1} entries to a vector with {-1,1} entries."""
    return np.where(
        binary_syndrome == 0,
        np.full(shape=binary_syndrome.shape, fill_value=1),
        np.full(shape=binary_syndrome.shape, fill_value=-1),
    )


def get_binary_from_analog(analog_syndrome: NDArray[np.float64]) -> NDArray[np.int32]:
    """Returns the thresholded binary vector.

    Since in {-1,+1} notation -1 indicates a check violation, we map values <= 0 to 1 and values > 0 to 0.
    """
    return np.where(analog_syndrome <= 0.0, 1, 0).astype(np.int32)


def save_results(
    success_cnt: int,
    nr_runs: int,
    p: float,
    s: float,
    input_vals: dict[str, Any],
    outfile: str,
    code_params: dict[str, int],
    bp_params: BpParams | None,
    err_side: str = "X",
    bp_iterations: int | None = None,
) -> dict[str, Any]:
    """Save results of a simulation run to a json file."""
    ler, ler_eb, wer, wer_eb = calculate_error_rates(success_cnt, nr_runs, code_params)

    output: dict[str, Any] = {
        "code_K": code_params["k"],
        "code_N": code_params["n"],
        "nr_runs": nr_runs,
        "pers": p,
        "sers": s,
        f"{err_side}_ler": ler,
        f"{err_side}_ler_eb": ler_eb,
        f"{err_side}_wer": wer,
        f"{err_side}_wer_eb": wer_eb,
        f"{err_side}_success_cnt": success_cnt,
        "avg_bp_iterations": bp_iterations / nr_runs if bp_iterations is not None else 0,
        "bp_params": bp_params,
    }

    output.update(input_vals)
    output["bias"] = replace_inf(output["bias"])
    with Path(outfile).open(mode="w", encoding=locale.getpreferredencoding(False)) as out:
        json.dump(output, out, ensure_ascii=False, indent=4, default=lambda o: o.__dict__)
    return output



"""This module contains the functions for the multiround decoding simulation."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from pymatching import Matching
from scipy.sparse import block_diag, csr_matrix, eye, hstack

from ..utils.simulation_utils import (
    get_virtual_check_init_vals,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray


def build_multiround_pcm(pcm: NDArray[np.int32], repetitions: int, matrix_format: str = "csr") -> csr_matrix:
    """Builds the multiround parity-check matrix as described in the paper.

    Each row corresponds to a round of measurements, the matrix for r repetitions has the form
    H3D := (H_diag | id_r), where H_diag is the r-diagonal matrix containing the parity-check matrix H and
    id_diag is a staircase block matrix containing two m-identity matrices in each row and column except the first one.
    """
    if not isinstance(pcm, csr_matrix):
        pcm = csr_matrix(pcm)

    pcm_rows, _ = pcm.shape

    # Construct the block of PCMs
    h_3dpcm = block_diag([pcm] * (repetitions + 1), format=matrix_format)

    # Construct the block of identity matrices
    h_3did_diag = block_diag([eye(pcm_rows, format=matrix_format)] * (repetitions + 1), format=matrix_format)

    # Construct the block of identity matrices
    h_3did_offdiag = eye(pcm_rows * (repetitions + 1), k=-pcm_rows, format=matrix_format)

    # Construct the block of identity matrices
    h_3did = h_3did_diag + h_3did_offdiag

    # # hstack the two blocks
    return hstack([h_3dpcm, h_3did], format=matrix_format)


def move_syndrome(syndrome: NDArray[Any], data_type: Any = np.int32) -> NDArray[Any]:  # noqa: ANN401
    """Slides the window one region up, i.e., the syndrome of the first half is overwritten by the second half."""
    region_size = int(syndrome.shape[1] / 2)  # number of rounds in each region, T

    # zero likes syndrome
    new_syndrome = np.zeros(syndrome.shape, dtype=data_type)

    # move the syndromes from the Tth round to the 0th round
    new_syndrome[:, :region_size] = syndrome[:, region_size:]
    return new_syndrome


def get_updated_decoder(
    decoding_method: str,
    decoder: Any,  # noqa: ANN401
    new_channel: NDArray[np.float64],
    h3d: Any | None = None,  # noqa: ANN401
) -> Any:  # noqa: ANN401
    """Updates the decoder with the new channel information and returns the updated decoder object."""
    if decoding_method == "bposd":
        decoder.update_channel_probs(new_channel)
        return decoder
    if decoding_method == "matching":
        weights = np.clip(
            np.log((1 - new_channel) / new_channel),
            a_min=-16777215,
            a_max=16777215,
        )
        return Matching(h3d, weights=weights)
    msg = "Unknown decoding method"
    raise ValueError(msg, decoding_method)


def decode_multiround(
    syndrome: NDArray[np.int32],
    pcm: NDArray[np.int32],
    decoder: Any,  # noqa: ANN401
    channel_probs: NDArray[np.float64],  # needed for matching decoder does not have an update weights method
    repetitions: int,
    analog_syndr: NDArray[np.float64] | None,
    last_round: bool = False,
    check_block_size: int = 0,
    sigma: float = 0.0,
    h3d: NDArray[np.int32] | None = None,  # needed for matching decoder
    decoding_method: str = "bposd",  # bposd or matching
) -> tuple[Any, NDArray[np.int32], NDArray[np.float64], int]:
    """Overlapping window decoding.

    First, we compute the difference syndrome from the recorded syndrome of each measurement round for all measurement
    rounds of the current window (consisting of two regions with equal size).
    Then, we apply the correction returned from the decoder on the first region (commit region).
    We then propagate the syndrome through the whole window (i.e., to the end of the second region).
    """
    analog_tg = analog_syndr is not None
    # convert syndrome to difference syndrome
    diff_syndrome = syndrome.copy()
    diff_syndrome[:, 1:] = (syndrome[:, 1:] - syndrome[:, :-1]) % 2
    bp_iter = 0

    region_size = repetitions // 2  # assumes repetitions is even

    if analog_tg:
        # If we have analog information, we use it to initialize the time-like syndrome nodes, which are defined
        # in the block of the H3D matrix after the diagonal H block.
        analog_init_vals = get_virtual_check_init_vals(analog_syndr.flatten("F"), sigma)  # type: ignore[union-attr]

        new_channel = np.hstack((channel_probs[:check_block_size], analog_init_vals))

        # in the last round, we have a perfect syndrome round to make sure we're in the codespace
        if last_round:
            new_channel[-pcm.shape[0] :] = 1e-15

        decoder = get_updated_decoder(decoding_method, decoder, new_channel, h3d)

    elif last_round:
        new_channel = np.copy(channel_probs)
        new_channel[-pcm.shape[0] :] = 1e-15

        decoder = get_updated_decoder(decoding_method, decoder, new_channel, h3d)

    decoded = decoder.decode(diff_syndrome.flatten("F"))

    if decoding_method == "bposd":
        bp_iter = decoder.iter

    # extract space correction, first repetitions * n entire
    space_correction = decoded[: pcm.shape[1] * repetitions].reshape((repetitions, pcm.shape[1])).T
    # extract time correction

    if last_round is False:
        # this corresponds to the decoding on the second block of the H3D matrix
        time_correction = decoded[pcm.shape[1] * repetitions :].reshape((repetitions, pcm.shape[0])).T

        # append time correction with zeros
        time_correction = np.hstack((time_correction, np.zeros((pcm.shape[0], 1), dtype=np.int32)))

        # correct only in the commit region
        decoded = (np.cumsum(space_correction, 1) % 2)[:, region_size - 1]

        #  get the syndrome according to the correction
        corr_syndrome = (pcm @ decoded) % 2

        # propagate the syndrome correction through the tentative region
        syndrome[:, region_size:] = ((syndrome[:, region_size:] + corr_syndrome[:, None]) % 2).astype(np.int32)

        # apply the time correction of round region_size - 1 to the syndrome at the beginning of the tentative region
        syndrome[:, region_size] = ((syndrome[:, region_size] + time_correction[:, region_size - 1]) % 2).astype(
            np.int32
        )

    else:
        # correct in the commit and tentative region as the last round stabilizer is perfect
        decoded = (np.cumsum(space_correction, 1) % 2)[:, -1]

    return decoded.astype(np.int32), syndrome, analog_syndr, bp_iter



"""Package for code construction."""

from __future__ import annotations

import json
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any

import ldpc.code_util
import numpy as np
import scipy.io as sio
import scipy.sparse as scs
from bposd.hgp import hgp
from ldpc import mod2
from scipy import sparse

if TYPE_CHECKING:
    from numpy.typing import NDArray


def is_all_zeros(array: NDArray[np.int32]) -> bool:
    """Check if array is all zeros."""
    return not np.any(array)


def run_checks_scipy(
    d_1: NDArray[np.int32], d_2: NDArray[np.int32], d_3: NDArray[np.int32], d_4: NDArray[np.int32]
) -> None:
    """Run checks on the boundary maps."""
    sd_1 = scs.csr_matrix(d_1)
    sd_2 = scs.csr_matrix(d_2)
    sd_3 = scs.csr_matrix(d_3)
    sd_4 = scs.csr_matrix(d_4)

    if not (
        is_all_zeros((sd_1 * sd_2).todense() % 2)
        and is_all_zeros((sd_2 * sd_3).todense() % 2)
        and is_all_zeros((sd_3 * sd_4).todense() % 2)
    ):
        msg = "Error generating 4D code, boundary maps do not square to zero"
        raise RuntimeError(msg)


def generate_4d_product_code(
    a_1: NDArray[np.int32], a_2: NDArray[np.int32], a_3: NDArray[np.int32], p: NDArray[np.int32], checks: bool = True
) -> tuple[NDArray[np.int32], NDArray[np.int32], NDArray[np.int32], NDArray[np.int32]]:
    """Generate 4D product code."""
    r, c = p.shape
    id_r: NDArray[np.int32] = np.identity(r, dtype=np.int32)
    id_c: NDArray[np.int32] = np.identity(c, dtype=np.int32)
    id_n0: NDArray[np.int32] = np.identity(a_1.shape[0], dtype=np.int32)
    id_n1: NDArray[np.int32] = np.identity(a_2.shape[0], dtype=np.int32)
    id_n2: NDArray[np.int32] = np.identity(a_3.shape[0], dtype=np.int32)
    id_n3: NDArray[np.int32] = np.identity(a_3.shape[1], dtype=np.int32)

    d_1: NDArray[np.int32] = np.hstack((np.kron(a_1, id_r), np.kron(id_n0, p))).astype(np.int32)

    x = np.hstack((np.kron(a_2, id_r), np.kron(id_n1, p)))
    y = np.kron(a_1, id_c)
    dims = (y.shape[0], x.shape[1] - y.shape[1])
    z = np.hstack((np.zeros(dims, dtype=np.int32), y))
    d_2 = np.vstack((x, z))

    x = np.hstack((np.kron(a_3, id_r), np.kron(id_n2, p)))
    y = np.kron(a_2, id_c)
    dims = (y.shape[0], x.shape[1] - y.shape[1])
    z = np.hstack((np.zeros(dims, dtype=np.int32), y))
    d_3: NDArray[np.int32] = np.vstack((x, z)).astype(np.int32)

    d_4: NDArray[np.int32] = np.vstack((np.kron(id_n3, p), np.kron(a_3, id_c)))

    if checks:
        run_checks_scipy(d_1, d_2, d_3, d_4)

    return d_1, d_2, d_3, d_4


def generate_3d_product_code(
    a_1: NDArray[np.int32], a_2: NDArray[np.int32], p: NDArray[np.int32]
) -> tuple[NDArray[np.int32], NDArray[np.int32], NDArray[np.int32]]:
    """Generate 3D product code."""
    r, c = p.shape
    id_r: NDArray[np.int32] = np.identity(r, dtype=np.int32)
    id_c: NDArray[np.int32] = np.identity(c, dtype=np.int32)
    id_n0: NDArray[np.int32] = np.identity(a_1.shape[0], dtype=np.int32)
    id_n1: NDArray[np.int32] = np.identity(a_2.shape[0], dtype=np.int32)
    id_n2: NDArray[np.int32] = np.identity(a_2.shape[1], dtype=np.int32)

    d_1: NDArray[np.int32] = np.hstack((np.kron(a_1, id_r), np.kron(id_n0, p)))

    x = np.hstack((np.kron(a_2, id_r), np.kron(id_n1, p)))
    y = np.kron(a_1, id_c)
    dims = (y.shape[0], x.shape[1] - y.shape[1])
    z = np.hstack((np.zeros(dims, dtype=np.int32), y))
    d_2: NDArray[np.int32] = np.vstack((x, z))

    d_3: NDArray[np.int32] = np.vstack((np.kron(id_n2, p), np.kron(a_2, id_c)))

    if not (is_all_zeros(d_1 @ d_2 % 2) and is_all_zeros(d_2 @ d_3 % 2)):
        msg = "Error generating 3D code, boundary maps do not square to zero"
        raise RuntimeError(msg)

    return d_1, d_2, d_3


def create_outpath(codename: str) -> str:
    """Create output path for code files."""
    path = f"generated_codes/{codename}/"
    Path(path).mkdir(parents=True, exist_ok=True)
    return path


def save_code(
    hx: NDArray[np.int32],
    hz: NDArray[np.int32],
    mx: NDArray[np.int32],
    mz: NDArray[np.int32],
    codename: str,
    lx: NDArray[np.int32] | None,
    lz: NDArray[np.int32] | None,
) -> None:
    """Save code to files."""
    path = create_outpath(codename)
    ms = [hx, hz, mx, mz, lx, lz] if lx is not None and lz is not None else [hx, hz, mx, mz]
    names: list[str] = ["hx", "hz", "mx", "mz", "lx", "lz"]
    for mat, name in zip(ms, names):
        if mat is not None:
            np.savetxt(path + name + ".txt", mat, fmt="%i")
            sio.mmwrite(
                path + name + ".mtx",
                sparse.coo_matrix(mat),
                comment="Field: GF(2)",
            )


def run_compute_distances(codename: str) -> None:
    """Run compute distances bash script."""
    path = "generated_codes/" + codename
    subprocess.run(["bash", "compute_distances.sh", path], check=False)  # noqa: S603, S607


def _compute_distances(hx: NDArray[np.int32], hz: NDArray[np.int32], codename: str) -> None:
    run_compute_distances(codename)
    code_dict: Any = {}
    _m, n = hx.shape
    code_k = n - mod2.rank(hx) - mod2.rank(hz)
    with Path(f"generated_codes/{codename}/info.txt").open() as f:
        code_dict = dict(
            line[: line.rfind("#")].split(" = ") for line in f if not line.startswith("#") and line.strip()
        )

    code_dict["n"] = n
    code_dict["k"] = code_k
    code_dict["dX"] = int(code_dict["dX"])
    code_dict["dZ"] = int(code_dict["dZ"])
    code_dict["dMX"] = int(code_dict["dMX"])
    code_dict["dMZ"] = int(code_dict["dMZ"])

    with Path(f"generated_codes/{codename}/code_params.txt").open("w") as file:
        file.write(json.dumps(code_dict))


def _compute_logicals(hx: NDArray[np.int32], hz: NDArray[np.int32]) -> tuple[NDArray[np.int32], NDArray[np.int32]]:
    def compute_lz(hx: NDArray[np.int32], hz: NDArray[np.int32]) -> NDArray[np.int32]:
        # lz logical operators
        # lz\in ker{hx} AND \notin Im(Hz.T)

        ker_hx = mod2.nullspace(hx)  # compute the kernel basis of hx
        im_hz_t = mod2.row_basis(hz)  # compute the image basis of hz.T

        # in the below we row reduce to find vectors in kx that are not in the image of hz.T.
        log_stack = np.vstack([im_hz_t, ker_hx])
        pivots = mod2.row_echelon(log_stack.T)[3]
        log_op_indices = [i for i in range(im_hz_t.shape[0], log_stack.shape[0]) if i in pivots]
        return log_stack[log_op_indices]

    lx = compute_lz(hz, hx)
    lz = compute_lz(hx, hz)
    return lx, lz


def create_code(
    constructor: str,
    seed_codes: list[NDArray[np.int32]],
    codename: str,
    compute_distance: bool = False,
    compute_logicals: bool = False,
    checks: bool = False,
) -> None:
    """Create 4D code."""
    # Construct initial 2 dim code
    if constructor == "hgp":
        code = hgp(seed_codes[0], seed_codes[1])
    else:
        msg = f"No constructor specified or the specified constructor {constructor} not implemented."
        raise ValueError(msg)

    # Extend to 3D HGP
    a1 = code.hx
    a2 = code.hz.T
    res = generate_3d_product_code(a1, a2, seed_codes[2])

    # Build 4D HGP code
    mx, hx, hz_t, mz_t = generate_4d_product_code(*res, seed_codes[3], checks=checks)

    hz = hz_t.T
    mz = mz_t.T

    # Perform checks
    if np.any(hz_t @ mz_t % 2) or np.any(hx @ hz_t % 2) or np.any(mx @ hx % 2):
        msg = "err"
        raise RuntimeError(msg)
    save_code(hx, hz, mx, mz, codename, lx=None, lz=None)

    if compute_logicals:
        lx, lz = _compute_logicals(hx, hz)
        save_code(hx, hz, mx, mz, codename, lx=lx, lz=lz)

    else:
        save_code(hx, hz, mx, mz, codename, lx=None, lz=None)

    if compute_distance:
        _compute_distances(hx, hz, codename)


if __name__ == "__main__":
    for d in range(3, 8):
        seed_codes = [
            ldpc.codes.ring_code(d),
            ldpc.codes.ring_code(d),
            ldpc.codes.ring_code(d),
            ldpc.codes.ring_code(d),
        ]

        constructor = "hgp"
        codename = f"4D_toric_{d:d}"
        compute_distance = False
        compute_logicals = True
        create_code(
            constructor,
            seed_codes,
            codename,
            compute_distance,
            compute_logicals,
        )



"""example simulation script"""

import numpy as np
import scipy
from matplotlib import pyplot as plt

from mqt.qecc.analog_information_decoding.simulators.analog_tannergraph_decoding import AtdSimulator
from mqt.qecc.analog_information_decoding.utils.data_utils import BpParams

code_path = "src/mqt/qecc/analog_information_decoding/codes/lifted_product/lp_l="
s = np.linspace(0.10, 0.4, 11)
p = 0.05
for bp_method in ["msl"]:
    for decoder in ["atd"]:
        for c in [16, 21, 30]:
            Hx = scipy.sparse.load_npz(code_path + str(c) + "_hx.txt").toarray()
            Hz = scipy.sparse.load_npz(code_path + str(c) + "_hz.txt").toarray()
            Lx = scipy.sparse.load_npz(code_path + str(c) + "_lx.txt").toarray()
            Lz = scipy.sparse.load_npz(code_path + str(c) + "_lz.txt").toarray()
            lers = []
            ebs = []
            for sigma in s:
                print(sigma)  # noqa: T201
                sim = AtdSimulator(
                    hx=Hx,
                    lx=Lx,
                    hz=Hz,
                    lz=Lz,
                    codename=str(c),
                    data_err_rate=p,
                    sigma=sigma,
                    seed=666,
                    bp_params=BpParams(
                        max_bp_iter=100,
                        bp_method=bp_method,
                        osd_order=10,
                        osd_method="osd_cs",
                        ms_scaling_factor=0.75,
                        cutoff=5.0,
                    ),
                    decoding_method=decoder,
                )
                out = sim.run(samples=1500)

                ler = (
                    out["z_ler"] * (1 - out["x_ler"]) + out["x_ler"] * (1 - out["z_ler"]) + out["x_ler"] * out["z_ler"]
                )
                ler_eb = (
                    out["z_ler_eb"] * (1 - out["x_ler_eb"])
                    + out["x_ler_eb"] * (1 - out["z_ler_eb"])
                    + out["x_ler_eb"] * out["z_ler_eb"]
                )
                lers.append(ler)
                ebs.append(ler_eb)

            plt.errorbar(
                s,
                lers,
                ebs,
                label=f"l={c}, {decoder} {bp_method}",
                marker="o",
                linestyle="solid",
            )

plt.legend()
plt.xlabel("sigma")
plt.ylabel("LER")
plt.yscale("log")
plt.show()



"""Simulation of Non-deterministic fault tolerant state preparation."""

from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import stim
from qiskit.converters import circuit_to_dag, dag_to_circuit

from ..codes import InvalidCSSCodeError

if TYPE_CHECKING:  # pragma: no cover
    import numpy.typing as npt
    from qiskit import QuantumCircuit

    from ..codes import CSSCode


class NoisyNDFTStatePrepSimulator:
    """Class for simulating noisy state preparation circuit using a depolarizing noise model."""

    def __init__(
        self,
        state_prep_circ: QuantumCircuit,
        code: CSSCode,
        p: float,
        zero_state: bool = True,
        parallel_gates: bool = True,
    ) -> None:
        """Initialize the simulator.

        Args:
            state_prep_circ: The state preparation circuit.
            code: The code to simulate.
            p: The error rate.
            zero_state: Whether thezero state is prepared or nor.
            parallel_gates: Whether to allow for parallel execution of gates.
        """
        if code.Hx is None or code.Hz is None:
            msg = "The code must have both X and Z checks."
            raise InvalidCSSCodeError(msg)

        self.circ = state_prep_circ
        self.num_qubits = state_prep_circ.num_qubits
        self.code = code
        self.p = p
        self.zero_state = zero_state
        # Store which measurements are X, Z or data measurements.
        # The indices refer to the indices of the measurements in the stim circuit.
        self.x_verification_measurements: list[int] = []
        self.z_verification_measurements: list[int] = []
        self.x_measurements: list[int] = []
        self.z_measurements: list[int] = []
        self.data_measurements: list[int] = []
        self.parallel_gates = parallel_gates
        self.n_measurements = 0
        self.stim_circ = stim.Circuit()
        self.decoder = LutDecoder(code)
        self.set_p(p)

    def set_p(self, p: float) -> None:
        """Set the error rate.

        This reinitializes the stim circuit.

        Args:
        p: The error rate.
        """
        self.x_verification_measurements = []
        self.z_verification_measurements = []
        self.x_measurements = []
        self.z_measurements = []
        self.data_measurements = []
        self.n_measurements = 0
        self.p = p
        self._reused_qubits = 0
        self.stim_circ = self.to_stim_circ()
        self.num_qubits = (
            self.stim_circ.num_qubits
            - (len(self.x_verification_measurements) + len(self.z_verification_measurements))
            + self._reused_qubits
        )
        self.measure_stabilizers()
        if self.zero_state:
            self.measure_z()
        else:
            self.measure_x()

    def to_stim_circ(self) -> stim.Circuit:
        """Convert a QuantumCircuit to a noisy STIM circuit.

        A depolarizing error model is used:
        - Single-qubit gates and idling qubits are followed by a single-qubit Pauli error with probability 2/9 p. This reflects the fact that two-qubit gates are more likely to fail.
        - Two-qubit gates are followed by a two-qubit Pauli error with probability p/15.
        - Measurements flip with a probability of 2/3 p.
        - Qubit are initialized in the -1 Eigenstate with probability 2/3 p.
        """
        initialized = [False for _ in self.circ.qubits]
        stim_circuit = stim.Circuit()
        ctrls = []

        def idle_error(used_qubits: list[int]) -> None:
            for q in self.circ.qubits:
                qubit = self.circ.find_bit(q)[0]
                if initialized[qubit] and qubit not in used_qubits:
                    stim_circuit.append_operation("DEPOLARIZE1", [self.circ.find_bit(q)[0]], [2 * self.p / 3])

        dag = circuit_to_dag(self.circ)
        layers = dag.layers()
        used_qubits: list[int] = []
        targets = set()
        measured: defaultdict[int, int] = defaultdict(int)
        for layer in layers:
            layer_circ = dag_to_circuit(layer["graph"])

            # Apply idling errors to all qubits that were unused in the previous layer
            if len(used_qubits) > 0:
                idle_error(used_qubits)

            used_qubits = []
            for gate in layer_circ.data:
                if gate.operation.name == "h":
                    qubit = self.circ.find_bit(gate.qubits[0])[0]
                    ctrls.append(qubit)
                    if initialized[qubit]:
                        stim_circuit.append_operation("H", [qubit])
                        stim_circuit.append_operation("DEPOLARIZE1", [qubit], [2 * self.p / 3])
                        if not self.parallel_gates:
                            idle_error([qubit])
                        else:
                            used_qubits.append(qubit)

                elif gate.operation.name == "cx":
                    ctrl = self.circ.find_bit(gate.qubits[0])[0]
                    target = self.circ.find_bit(gate.qubits[1])[0]
                    targets.add(target)
                    if not initialized[ctrl]:
                        if ctrl in ctrls:
                            stim_circuit.append_operation("H", [ctrl])
                            stim_circuit.append_operation("Z_ERROR", [ctrl], [2 * self.p / 3])  # Wrong initialization
                        else:
                            stim_circuit.append_operation("X_ERROR", [ctrl], [2 * self.p / 3])  # Wrong initialization
                        initialized[ctrl] = True
                    if not initialized[target]:
                        stim_circuit.append_operation("X_ERROR", [target], [2 * self.p / 3])  # Wrong initialization
                        initialized[target] = True

                    stim_circuit.append_operation("CX", [ctrl, target])
                    stim_circuit.append_operation("DEPOLARIZE2", [ctrl, target], [self.p])
                    if not self.parallel_gates:
                        idle_error([ctrl, target])
                    else:
                        used_qubits.extend([ctrl, target])

                elif gate.operation.name == "measure":
                    anc = self.circ.find_bit(gate.qubits[0])[0]
                    stim_circuit.append_operation("X_ERROR", [anc], [2 * self.p / 3])
                    stim_circuit.append_operation("MR", [anc])
                    if anc in targets:
                        self.z_verification_measurements.append(self.n_measurements)
                    else:
                        self.x_verification_measurements.append(self.n_measurements)
                    self.n_measurements += 1
                    if not self.parallel_gates:
                        idle_error([anc])
                    else:
                        used_qubits.append(anc)
                    initialized[anc] = False
                    measured[anc] += 1
                    if measured[anc] == 2:
                        self._reused_qubits += 1

        return stim_circuit

    def measure_stabilizers(self) -> None:
        """Measure the stabilizers of the code.

        An ancilla is used for each measurement.
        """
        assert self.code.Hx is not None
        assert self.code.Hz is not None

        for check in self.code.Hx:
            supp = _support(check)
            anc = self.stim_circ.num_qubits
            self.stim_circ.append_operation("H", [anc])
            for q in supp:
                self.stim_circ.append_operation("CX", [anc, q])
            self.stim_circ.append_operation("MRX", [anc])
            self.x_measurements.append(self.n_measurements)
            self.n_measurements += 1

        for check in self.code.Hz:
            supp = _support(check)
            anc = self.stim_circ.num_qubits
            for q in supp:
                self.stim_circ.append_operation("CX", [q, anc])
            self.stim_circ.append_operation("MRZ", [anc])
            self.z_measurements.append(self.n_measurements)
            self.n_measurements += 1

    def measure_z(self) -> None:
        """Measure all data qubits in the Z basis."""
        self.data_measurements = [self.n_measurements + i for i in range(self.num_qubits)]
        self.n_measurements += self.num_qubits
        self.stim_circ.append_operation("MRZ", list(range(self.num_qubits)))

    def measure_x(self) -> None:
        """Measure all data qubits in the X basis."""
        self.data_measurements = [self.n_measurements + i for i in range(self.num_qubits)]
        self.n_measurements += self.num_qubits
        self.stim_circ.append_operation("MRX", list(range(self.num_qubits)))

    def logical_error_rate(
        self,
        shots: int = 100000,
        shots_per_batch: int = 100000,
        at_least_min_errors: bool = True,
        min_errors: int = 500,
    ) -> tuple[float, float, int, int]:
        """Estimate the logical error rate of the code.

        Args:
            shots: The number of shots to use.
            shots_per_batch: The number of shots per batch.
            at_least_min_errors: Whether to continue simulating until at least min_errors are found.
            min_errors: The minimum number of errors to find before stopping.
        """
        batch = min(shots_per_batch, shots)
        p_l = 0.0
        r_a = 0.0

        num_logical_errors = 0

        if self.zero_state:
            self.decoder.generate_x_lut()
        else:
            self.decoder.generate_z_lut()

        i = 1
        while i <= int(np.ceil(shots / batch)) or at_least_min_errors:
            num_logical_errors_batch, discarded_batch = self._simulate_batch(batch)

            logging.log(
                logging.INFO,
                f"Batch {i}: {num_logical_errors_batch} logical errors and {discarded_batch} discarded shots. {batch - discarded_batch} shots used.",
            )
            p_l_batch = num_logical_errors_batch / (batch - discarded_batch) if discarded_batch != batch else 0.0
            p_l = ((i - 1) * p_l + p_l_batch) / i

            r_a_batch = 1 - discarded_batch / batch

            # Update statistics
            num_logical_errors += num_logical_errors_batch
            r_a = ((i - 1) * r_a + r_a_batch) / i

            if at_least_min_errors and num_logical_errors >= min_errors:
                break
            i += 1

        return p_l / self.code.k, r_a, num_logical_errors, i * batch

    def _simulate_batch(self, shots: int = 1024) -> tuple[int, int]:
        sampler = self.stim_circ.compile_sampler()
        detection_events = sampler.sample(shots)

        # Filter events where the verification circuit flagged
        verification_measurements = self.x_verification_measurements + self.z_verification_measurements
        index_array = np.where(np.all(detection_events[:, verification_measurements] == 0, axis=1))[0]
        filtered_events = detection_events[index_array].astype(np.int8)

        if len(filtered_events) == 0:  # All events were discarded
            return 0, shots

        state = filtered_events[:, self.data_measurements]

        if self.zero_state:
            checks = filtered_events[:, self.z_measurements]
            observables = self.code.Lz
            estimates = self.decoder.batch_decode_x(checks)
        else:
            checks = filtered_events[:, self.x_measurements]
            observables = self.code.Lx
            estimates = self.decoder.batch_decode_z(checks)

        corrected = state + estimates

        num_discarded = detection_events.shape[0] - filtered_events.shape[0]
        num_logical_errors: int = np.sum(
            np.any(corrected @ observables.T % 2 != 0, axis=1)
        )  # number of non-commuting corrected states
        return num_logical_errors, num_discarded

    def plot_state_prep(self, ps: list[float], min_errors: int = 500, name: str | None = None) -> None:
        """Plot the logical error rate and accaptence rate as a function of the physical error rate.

        Args:
            ps: The physical error rates to plot.
            min_errors: The minimum number of errors to find before stopping.
            name: The name of the plot.
        """
        p_ls = []
        r_as = []
        for p in ps:
            self.set_p(p)
            p_l, r_a, _num_logical_errors, _num_shots = self.logical_error_rate(min_errors=min_errors)
            p_ls.append(p_l)
            r_as.append(r_a)

        plt.subplot(1, 2, 1)
        plt.plot(ps, p_ls, marker="o", label=name)
        plt.xscale("log")
        plt.yscale("log")
        plt.xlabel("Physical error rate")
        plt.ylabel("Logical error rate")

        if name is not None:
            plt.legend()

        plt.subplot(1, 2, 2)
        plt.plot(ps, r_as, marker="o", label=name)
        plt.xscale("log")
        plt.yscale("log")
        plt.xlabel("Physical error rate")
        plt.ylabel("Acceptance rate")
        if name is not None:
            plt.legend()
        plt.tight_layout()


class LutDecoder:
    """Lookup table decoder for a CSSCode."""

    def __init__(self, code: CSSCode, init_luts: bool = True) -> None:
        """Initialize the decoder.

        Args:
            code: The code to decode.
            init_luts: Whether to initialize the lookup tables at object creation.
        """
        self.code = code
        self.x_lut: dict[bytes, npt.NDArray[np.int8]] = {}
        self.z_lut: dict[bytes, npt.NDArray[np.int8]] = {}
        if init_luts:
            self.generate_x_lut()
            self.generate_z_lut()

    def batch_decode_x(self, syndromes: npt.NDArray[np.int_]) -> npt.NDArray[np.int8]:
        """Decode the X errors given a batch of syndromes."""
        return np.apply_along_axis(self.decode_x, 1, syndromes)

    def batch_decode_z(self, syndromes: npt.NDArray[np.int_]) -> npt.NDArray[np.int8]:
        """Decode the Z errors given a batch of syndromes."""
        return np.apply_along_axis(self.decode_z, 1, syndromes)

    def decode_x(self, syndrome: npt.NDArray[np.int8]) -> npt.NDArray[np.int8]:
        """Decode the X errors given a syndrome."""
        if len(self.x_lut) == 0:
            self.generate_x_lut()
        return self.x_lut[syndrome.tobytes()]

    def decode_z(self, syndrome: npt.NDArray[np.int8]) -> npt.NDArray[np.int8]:
        """Decode the Z errors given a syndrome."""
        if len(self.z_lut) == 0:
            self.generate_z_lut()
        return self.z_lut[syndrome.tobytes()]

    def generate_x_lut(self) -> None:
        """Generate the lookup table for the X errors."""
        if len(self.x_lut) != 0:
            return

        assert self.code.Hz is not None, "The code does not have a Z stabilizer matrix."
        self.x_lut = LutDecoder._generate_lut(self.code.Hz)
        if self.code.is_self_dual():
            self.z_lut = self.x_lut

    def generate_z_lut(self) -> None:
        """Generate the lookup table for the Z errors."""
        if len(self.z_lut) != 0:
            return

        assert self.code.Hx is not None, "The code does not have an X stabilizer matrix."
        self.z_lut = LutDecoder._generate_lut(self.code.Hx)
        if self.code.is_self_dual():
            self.z_lut = self.x_lut

    @staticmethod
    def _generate_lut(checks: npt.NDArray[np.int_]) -> dict[bytes, npt.NDArray[np.int_]]:
        """Generate a lookup table for the stabilizer state.

        The lookup table maps error syndromes to their errors.
        """
        n_qubits = checks.shape[1]

        syndromes = defaultdict(list)
        lut: dict[bytes, npt.NDArray[np.int8]] = {}
        for i in range(2**n_qubits):
            state: npt.NDArray[np.int_] = np.array(list(np.binary_repr(i).zfill(n_qubits))).astype(np.int8)
            syndrome = checks @ state % 2
            syndromes[syndrome.astype(np.int8).tobytes()].append(state)

        # Sort according to weight
        for key, v in syndromes.items():
            lut[key] = np.array(min(v, key=np.sum))

        return lut


def _support(v: npt.NDArray[np.int_]) -> npt.NDArray[np.int_]:
    """Return the support of a vector."""
    return np.where(v != 0)[0]



"""Analog Tannergraph Decoding Simulator."""

from __future__ import annotations

import json
import locale
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
from ldpc import bposd_decoder

from ..utils import simulation_utils
from ..utils.data_utils import (
    BpParams,
    calculate_error_rates,
    is_converged,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray


class AnalogTannergraphDecoder:
    """Analog Tannergraph decoder.

    Builds the analog tanner graph and uses this as decoding graph.
    """

    def __init__(
        self,
        pcm: NDArray[np.int32],
        bp_params: BpParams,
        error_channel: NDArray[np.float64],
        sigma: float = 0.0,
        ser: float | None = None,
    ) -> None:
        """Initialize the decoder."""
        self.m, self.n = pcm.shape
        self.sigma = sigma
        self.H = pcm
        self.bp_params = bp_params
        self.atg = get_analog_pcm(pcm)
        self.syndr_err_rate = ser
        self.error_channel = error_channel

        if self.sigma <= 0.0:
            if self.syndr_err_rate is None:
                msg = "Either sigma or ser must be specified"
                raise ValueError(msg)
            self.sigma = simulation_utils.get_sigma_from_syndr_er(self.syndr_err_rate)
        elif self.syndr_err_rate is not None:
            msg = "Only one of sigma or ser must be specified"
            raise ValueError(msg)

        self.bposd_decoder = bposd_decoder(
            parity_check_matrix=self.atg,
            channel_probs=np.hstack((self.error_channel, np.zeros(self.m))),  # initd as dummy for now
            max_iter=self.bp_params.max_bp_iter,
            bp_method=self.bp_params.bp_method,
            osd_order=self.bp_params.osd_order,
            osd_method=self.bp_params.osd_method,
            ms_scaling_factor=self.bp_params.ms_scaling_factor,
            schedule=self.bp_params.schedule,
            omp_thread_count=self.bp_params.omp_thread_count,
            random_serial_schedule=self.bp_params.random_serial_schedule,
            serial_schedule_order=self.bp_params.serial_schedule_order,
        )

        # self.bp_decoder = bp_decoder(
        #     parity_check_matrix=self.atg,
        #     channel_probs=np.hstack((self.error_channel, np.zeros(self.m))),  # initd as dummy for now
        #     max_iter=self.bp_params.max_bp_iter,
        #     bp_method=self.bp_params.bp_method,
        #     ms_scaling_factor=self.bp_params.ms_scaling_factor,
        # )

    def _set_analog_syndrome(self, analog_syndrome: NDArray[np.float64]) -> None:
        """Initializes the error channel of the BP decoder.

        Ensures that the virtual nodes are initialized with the analog syndrome LLRs on decoding initialization.
        :param analog_syndrome: the analog syndrome values to initialize the virtual nodes with.
        """
        new_channel = np.hstack((
            self.bposd_decoder.channel_probs[: self.n],
            simulation_utils.get_virtual_check_init_vals(analog_syndrome, self.sigma),
        ))
        self.bposd_decoder.update_channel_probs(new_channel)

    def decode(self, analog_syndrome: NDArray[np.float64]) -> NDArray[np.int32]:
        """Decode a given analog syndrome."""
        self._set_analog_syndrome(analog_syndrome)
        return self.bposd_decoder.decode(simulation_utils.get_binary_from_analog(analog_syndrome))


class AtdSimulator:
    """Analog Tanner graph Decoding Simulator."""

    def __init__(
        self,
        hx: NDArray[np.int32],
        lx: NDArray[np.int32],
        hz: NDArray[np.int32],
        lz: NDArray[np.int32],
        codename: str,
        seed: int,
        bp_params: BpParams,
        data_err_rate: float,
        code_params: dict[str, Any],
        syndr_err_rate: float | None = None,
        sigma: float | None = None,
        bias: NDArray[np.float64] | None = None,
        experiment: str = "atd",
        decoding_method: str = "atd",
        output_path: str = "./",
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize the simulator."""
        if bias is None:
            bias = np.array([1.0, 1.0, 1.0])
        simulation_utils.set_seed(seed)
        self.Hx = hx
        self.Lx = lx
        self.Hz = hz
        self.Lz = lz
        self.codename = codename
        self.data_err_rate = data_err_rate
        self.bias = bias
        self.experiment = experiment
        self.decoding_method = decoding_method
        self.sigma = sigma
        self.outfile = output_path
        if sigma is None:
            if syndr_err_rate is None:
                msg = "Either sigma or ser must be specified"
                raise ValueError(msg)

            self.syndr_err_rate = syndr_err_rate
            synd_err_channel = simulation_utils.error_channel_setup(
                error_rate=self.syndr_err_rate,
                xyz_error_bias=self.bias,
                nr_qubits=1,
            )

        else:
            if syndr_err_rate is not None:
                msg = "Only one of sigma or ser must be specified"
                raise ValueError(msg)

            self.syndr_err_rate = simulation_utils.get_error_rate_from_sigma(sigma)
            synd_err_channel = simulation_utils.error_channel_setup(
                error_rate=self.syndr_err_rate,
                xyz_error_bias=self.bias,
                nr_qubits=1,
            )

        x_synd_err_rate = synd_err_channel[0][0] + synd_err_channel[1][0]  # x + y errors, 1st bit only
        z_synd_err_rate = synd_err_channel[2][0] + synd_err_channel[1][0]  # z + y errors, 1st bit only
        self.x_sigma = simulation_utils.get_sigma_from_syndr_er(x_synd_err_rate)
        self.z_sigma = simulation_utils.get_sigma_from_syndr_er(z_synd_err_rate)

        self.bp_params = bp_params
        self.save_interval = kwargs.get("save_interval", 1_000)
        self.eb_precision = kwargs.get("eb_precision", 1e-1)
        self.input_values = self.__dict__.copy()

        self.n = hx.shape[1]
        self.code_params = code_params
        del self.input_values["Hx"]
        del self.input_values["Lx"]
        del self.input_values["Hz"]
        del self.input_values["Lz"]

        # setup decoders
        if self.decoding_method == "atd":
            Decoder = AnalogTannergraphDecoder  # noqa: N806

        # single-sided error only, no bias
        self.full_error_channel = simulation_utils.error_channel_setup(
            error_rate=self.data_err_rate,
            xyz_error_bias=self.bias,
            nr_qubits=self.n,
        )
        self.x_decoder = Decoder(
            error_channel=self.full_error_channel[0] + self.full_error_channel[1],  # x + y errors
            pcm=self.Hz,
            sigma=self.x_sigma,
            bp_params=self.bp_params,
        )
        self.z_decoder = Decoder(
            error_channel=self.full_error_channel[2] + self.full_error_channel[1],  # z + y errors
            pcm=self.Hx,
            sigma=self.z_sigma,
            bp_params=self.bp_params,
        )
        self.x_bp_iterations = 0
        self.z_bp_iterations = 0

    def single_sample(self) -> tuple[bool, bool]:
        """Samples and error and decodes once. Returns if the decoding round was successful separately for each side."""
        residual_err: list[NDArray[np.int32]] = [
            np.zeros(self.n).astype(np.int32),
            np.zeros(self.n).astype(np.int32),
        ]  # no residual error
        (
            x_err,
            z_err,
        ) = simulation_utils.generate_err(  # no residual error, only one side needed
            nr_qubits=self.n,
            channel_probs=self.full_error_channel,
            residual_err=residual_err,
        )

        x_perf_syndr = (self.Hz @ x_err) % 2
        x_noisy_syndr = simulation_utils.get_noisy_analog_syndrome(sigma=self.x_sigma, perfect_syndr=x_perf_syndr)
        x_decoding = self.x_decoder.decode(x_noisy_syndr)[: self.n]
        self.x_bp_iterations += self.x_decoder.bposd_decoder.iter
        x_residual = (x_err + x_decoding) % 2

        z_perf_syndr = (self.Hx @ z_err) % 2
        z_noisy_syndr = simulation_utils.get_noisy_analog_syndrome(sigma=self.z_sigma, perfect_syndr=z_perf_syndr)
        z_decoding = self.z_decoder.decode(z_noisy_syndr)[: self.n]
        self.z_bp_iterations += self.z_decoder.bposd_decoder.iter
        z_residual = (z_err + z_decoding) % 2

        return not simulation_utils.is_logical_err(self.Lz, x_residual), not simulation_utils.is_logical_err(
            self.Lx, z_residual
        )

    def run(self, samples: int) -> dict[str, Any]:
        """Run the simulation."""
        x_success_cnt = 0
        z_success_cnt = 0
        for runs in range(1, samples + 1):
            out = self.single_sample()

            x_success_cnt += int(out[0])
            z_success_cnt += int(out[1])
            if runs % self.save_interval == 1:
                self.save_results(x_success_cnt, z_success_cnt, runs)

                # check convergence only once during each save interval
                if is_converged(
                    x_success_cnt,
                    z_success_cnt,
                    runs,
                    self.code_params,
                    self.eb_precision,
                ):
                    print("Result has converged.")  # noqa: T201
                    break

        x_ler, x_ler_eb, x_wer, x_wer_eb = calculate_error_rates(x_success_cnt, runs, self.code_params)
        z_ler, z_ler_eb, z_wer, z_wer_eb = calculate_error_rates(z_success_cnt, runs, self.code_params)
        output = {
            "code_K": self.code_params["k"],
            "code_N": self.code_params["n"],
            "nr_runs": runs,
            "pers": self.data_err_rate,
            "x_sigma": self.x_sigma,
            "z_sigma": self.z_sigma,
            "x_ler": x_ler,
            "x_ler_eb": x_ler_eb,
            "x_wer": x_wer,
            "x_wer_eb": x_wer_eb,
            "x_success_cnt": x_success_cnt,
            "z_ler": z_ler,
            "z_ler_eb": z_ler_eb,
            "z_wer": z_wer,
            "z_wer_eb": z_wer_eb,
            "z_success_cnt": z_success_cnt,
            "bp_params": self.bp_params,
            "x_bp_iterations": self.x_bp_iterations / runs,
            "z_bp_iterations": self.z_bp_iterations / runs,
        }

        self.save_results(x_success_cnt, z_success_cnt, runs)  # save final results

        return output

    def save_results(self, x_success_cnt: int, z_success_cnt: int, runs: int) -> dict[str, Any]:
        """Compute error rates and error bars and save output dict."""
        x_ler, x_ler_eb, x_wer, x_wer_eb = calculate_error_rates(x_success_cnt, runs, self.code_params)
        z_ler, z_ler_eb, z_wer, z_wer_eb = calculate_error_rates(z_success_cnt, runs, self.code_params)
        output = {
            "code_K": self.code_params["k"],
            "code_N": self.code_params["n"],
            "nr_runs": runs,
            "pers": self.data_err_rate,
            "x_sigma": self.x_sigma,
            "z_sigma": self.z_sigma,
            "x_ler": x_ler,
            "x_ler_eb": x_ler_eb,
            "x_wer": x_wer,
            "x_wer_eb": x_wer_eb,
            "x_success_cnt": x_success_cnt,
            "z_ler": z_ler,
            "z_ler_eb": z_ler_eb,
            "z_wer": z_wer,
            "z_wer_eb": z_wer_eb,
            "z_success_cnt": z_success_cnt,
            "bp_params": self.bp_params,
            "x_bp_iterations": self.x_bp_iterations / runs,
            "z_bp_iterations": self.z_bp_iterations / runs,
        }

        output.update(self.input_values)
        with Path(self.outfile).open(mode="w", encoding=locale.getpreferredencoding(False)) as f:
            json.dump(output, f, ensure_ascii=False, indent=4, default=lambda o: o.__dict__)
        return output


def get_analog_pcm(pcm: NDArray[np.int32]) -> NDArray[np.int32]:
    """Constructs apcm = [H | I_m] where I_m is the m x m identity matrix."""
    return np.hstack((pcm, np.identity(pcm.shape[0], dtype=np.int32)))



"""Quasi single shot simulator."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from bposd import bposd_decoder
from pymatching import Matching

from ..utils.data_utils import BpParams, _check_convergence
from ..utils.simulation_utils import (
    error_channel_setup,
    generate_err,
    generate_syndr_err,
    get_binary_from_analog,
    get_noisy_analog_syndrome,
    get_sigma_from_syndr_er,
    is_logical_err,
    save_results,
    set_seed,
)
from .memory_experiment_v2 import (
    build_multiround_pcm,
    decode_multiround,
    move_syndrome,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray


class QssSimulator:
    """Quasi single shot simulator."""

    def __init__(
        self,
        pcm: NDArray[np.int_],
        per: float,
        ser: float,
        logicals: NDArray[np.int_],
        bias: NDArray[np.float64],
        codename: str,
        bp_params: BpParams,
        code_params: dict[str, int],
        decoding_method: str = "bposd",  # bposd or matching
        check_side: str = "X",
        seed: int = 666,
        analog_tg: bool = False,
        repetitions: int = 0,
        rounds: int = 0,
        experiment: str = "qss",
        outpath: str = "/",
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize QSS Simulator.

        :param pcm: parity-check matrix of code.

        :param per: physical data error rate
        :param ser: syndrome error rate
        :param logicals: logical matrix
        :param bias: bias array
        :param codename: name of the code
        :param bp_params: BP decoder parameters
        :param check_side: side of the check (X or Z)
        :param seed: random seed
        :param analog_tg: switch analog decoding on/off
        :param repetitions: number of total syndrome measurements, i.e., total time steps. Must be even.
        :param rounds: number of decoding runs, i.e., number of times we slide the window - 1
        :param experiment: name of experiment, for outpath creation
        :param kwargs:
        """
        self.H = pcm
        self.data_err_rate = per
        self.syndr_err_rate = ser
        self.check_side = check_side
        self.L = logicals
        self.bias = bias
        self.codename = codename
        self.bp_params = bp_params
        self.decoding_method = decoding_method
        self.save_interval = kwargs.get("save_interval", 50)
        self.eb_precision = kwargs.get("eb_precision", 1e-2)
        self.analog_tg = analog_tg
        self.repetitions = repetitions
        if repetitions % 2 != 0:
            msg = "repetitions must be even"
            raise ValueError(msg)

        if self.decoding_method not in {"bposd", "matching"}:
            msg = "Decoding method must be either bposd or matching"
            raise ValueError(msg)

        if self.repetitions % 2 != 0:
            msg = "Repetitions must be even!"
            raise ValueError(msg)

        self.rounds = rounds
        self.experiment = experiment
        set_seed(seed)
        self.code_params = code_params
        self.input_values = self.__dict__.copy()

        self.outfile = outpath

        # Remove Arrays
        del self.input_values["H"]
        del self.input_values["L"]

        self.num_checks, self.num_qubits = self.H.shape

        self.x_bit_chnl, self.y_bit_chnl, self.z_bit_chnl = error_channel_setup(
            error_rate=self.data_err_rate,
            xyz_error_bias=bias,
            nr_qubits=self.num_qubits,
        )
        self.x_syndr_err_chnl, self.y_syndr_err_chnl, self.z_syndr_err_chnl = error_channel_setup(
            error_rate=self.syndr_err_rate,
            xyz_error_bias=bias,
            nr_qubits=self.num_checks,
        )
        if self.check_side == "X":
            self.err_idx = 1
            # Z bit/syndrome errors
            self.data_err_channel = self.y_bit_chnl + self.z_bit_chnl
            self.syndr_err_channel = 1.0 * (self.z_syndr_err_chnl + self.y_syndr_err_chnl)
        else:
            # we have X errors on qubits
            self.err_idx = 0
            # X bit/syndrome errors
            self.data_err_channel = self.x_bit_chnl + self.y_bit_chnl
            self.syndr_err_channel = 1.0 * (self.x_syndr_err_chnl + self.y_syndr_err_chnl)

        # initialize the multiround parity-check matrix as described in the paper
        self.H3D = build_multiround_pcm(
            self.H,
            self.repetitions - 1,
        )

        # the number of columns of the diagonal check matrix of the H3D matrix
        self.check_block_size = self.num_qubits * (self.repetitions)

        channel_probs: NDArray[np.float64] = np.zeros(self.H3D.shape[1]).astype(np.float64)
        # The bits corresponding to the columns of the diagonal H-bock of H3D are initialized with the bit channel
        channel_probs[: self.check_block_size] = np.array(self.data_err_channel.tolist() * (self.repetitions))

        # The remaining bits (corresponding to the identity block of H3D)
        # are initialized with the syndrome error channel
        channel_probs[self.check_block_size :] = np.array(self.syndr_err_channel.tolist() * (self.repetitions))

        # If we do ATG decoding, initialize sigma (syndrome noise strength)
        if self.analog_tg:
            self.sigma = get_sigma_from_syndr_er(
                self.syndr_err_channel[0]  # x/z + y
            )  # assumes all sigmas are the same
        self.bp_iterations = 0
        if self.decoding_method == "bposd":
            self.decoder = bposd_decoder(
                parity_check_matrix=self.H3D,
                channel_probs=channel_probs,
                max_iter=self.bp_params.max_bp_iter,
                bp_method=self.bp_params.bp_method,
                osd_order=self.bp_params.osd_order,
                osd_method=self.bp_params.osd_method,
                ms_scaling_factor=self.bp_params.ms_scaling_factor,
            )
        elif self.decoding_method == "matching":
            weights = np.log((1 - channel_probs) / channel_probs)
            self.decoder = Matching(self.H3D, weights=weights)
        self.channel_probs = channel_probs

    def _decode_multiround(
        self,
        syndrome_mat: NDArray[np.int_],
        analog_syndr_mat: NDArray[np.float64],
        last_round: bool = False,
    ) -> tuple[Any, NDArray[np.int32], NDArray[np.float64], int]:
        return decode_multiround(
            syndrome=syndrome_mat,
            pcm=self.H,
            decoder=self.decoder,
            repetitions=self.repetitions,
            last_round=last_round,
            analog_syndr=analog_syndr_mat,
            check_block_size=self.check_block_size,
            sigma=self.sigma,
            h3d=self.H3D if self.decoding_method == "matching" else None,  # avoid passing matrix in case not needed
            channel_probs=self.channel_probs,
            decoding_method=self.decoding_method,
        )

    def _single_sample(self) -> int:
        # prepare fresh syndrome matrix and error vector
        # each column == measurement result of a single timestep
        syndrome_mat: NDArray[np.int32] = np.zeros((self.num_checks, self.repetitions), dtype=np.int32)

        if self.analog_tg:
            analog_syndr_mat: NDArray[np.float64] = np.zeros((self.num_checks, self.repetitions), dtype=np.float64)

        err: NDArray[np.int32] = np.zeros(self.num_qubits, dtype=np.int32)
        cnt = 0  # counter for syndrome_mat

        for rnd in range(self.rounds):
            residual_err = [np.copy(err), np.copy(err)]
            err = generate_err(
                nr_qubits=self.num_qubits,
                channel_probs=(
                    self.x_bit_chnl,
                    self.y_bit_chnl,
                    self.z_bit_chnl,
                ),
                residual_err=residual_err,
            )[self.err_idx]  # only first or last vector needed, depending on side (X or Z)
            noiseless_syndrome = (self.H @ err) % 2

            # add syndrome error
            if rnd != (self.rounds - 1):
                if self.analog_tg:
                    analog_syndrome = get_noisy_analog_syndrome(noiseless_syndrome, self.sigma)
                    syndrome = get_binary_from_analog(analog_syndrome)
                else:
                    syndrome_error = generate_syndr_err(self.syndr_err_channel)
                    syndrome = (noiseless_syndrome + syndrome_error) % 2
            else:  # last round is perfect
                syndrome = np.copy(noiseless_syndrome)
                analog_syndrome = get_noisy_analog_syndrome(noiseless_syndrome, 0.0)  # no noise

            # fill the corresponding column of the syndrome/analog syndrome matrix
            syndrome_mat[:, cnt] += syndrome
            if self.analog_tg:
                analog_syndr_mat[:, cnt] += analog_syndrome

            cnt += 1  # move to next column of syndrome matrix

            if cnt == self.repetitions:  # if we have filled the syndrome matrix, decode
                if rnd != (self.rounds - 1):  # if not last round, decode and move syndrome
                    cnt = self.repetitions // 2  # reset counter to start of tentative region

                    # the correction is only the correction of the commit region
                    (corr, syndrome_mat, analog_syndr_mat, bp_iters) = self._decode_multiround(
                        syndrome_mat,
                        analog_syndr_mat,
                        last_round=False,
                    )
                    # we compute the average for all rounds since this equals a single sample
                    self.bp_iterations += int(bp_iters / self.rounds)
                    err = (err + corr) % 2
                    syndrome_mat = move_syndrome(syndrome_mat)
                    if self.analog_tg:
                        analog_syndr_mat = move_syndrome(analog_syndr_mat, data_type=np.float64)

                else:  # if we are in the last round, decode and stop
                    # the correction is the correction of the commit and tentative region
                    (corr, syndrome_mat, analog_syndr_mat, bp_iters) = self._decode_multiround(
                        syndrome_mat,
                        analog_syndr_mat,
                        last_round=True,
                    )
                    self.bp_iterations += int(bp_iters / self.rounds)
                    err = (err + corr) % 2
        return int(not is_logical_err(self.L, err))

    def _save_results(self, success_cnt: int, samples: int) -> dict[str, Any]:
        return save_results(
            success_cnt=success_cnt,
            nr_runs=samples,
            p=self.data_err_rate,
            s=self.syndr_err_rate,
            input_vals=self.input_values,
            outfile=self.outfile,
            code_params=self.code_params,
            err_side="z" if self.check_side == "X" else "x",
            bp_iterations=self.bp_iterations,
            bp_params=self.bp_params,
        )

    def run(self, samples: int = 1) -> dict[str, Any]:
        """Returns single data point."""
        success_cnt = 0
        for run in range(1, samples + 1):
            success_cnt += self._single_sample()
            if run % self.save_interval == 1:
                self._save_results(success_cnt, run)
                if _check_convergence(success_cnt, run, self.code_params, self.eb_precision):
                    print("Converged")  # noqa: T201
                    break
        return self._save_results(success_cnt, run)



"""Class for representing quantum error correction codes."""

from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from ldpc import mod2

if TYPE_CHECKING:  # pragma: no cover
    import numpy.typing as npt


class CSSCode:
    """A class for representing CSS codes."""

    def __init__(
        self,
        distance: int,
        Hx: npt.NDArray[np.int8] | None = None,  # noqa: N803
        Hz: npt.NDArray[np.int8] | None = None,  # noqa: N803
        x_distance: int | None = None,
        z_distance: int | None = None,
    ) -> None:
        """Initialize the code."""
        self.distance = distance
        self.x_distance = x_distance if x_distance is not None else distance
        self.z_distance = z_distance if z_distance is not None else distance

        if self.distance < 0:
            msg = "The distance must be a non-negative integer"
            raise InvalidCSSCodeError(msg)
        if Hx is None and Hz is None:
            msg = "At least one of the check matrices must be provided"
            raise InvalidCSSCodeError(msg)
        if self.x_distance < self.distance or self.z_distance < self.distance:
            msg = "The x and z distances must be greater than or equal to the distance"
            raise InvalidCSSCodeError(msg)
        if Hx is not None and Hz is not None:
            if Hx.shape[1] != Hz.shape[1]:
                msg = "Check matrices must have the same number of columns"
                raise InvalidCSSCodeError(msg)
            if np.any(Hx @ Hz.T % 2 != 0):
                msg = "The check matrices must be orthogonal"
                raise InvalidCSSCodeError(msg)

        self.Hx = Hx
        self.Hz = Hz
        self.n = Hx.shape[1] if Hx is not None else Hz.shape[1]  # type: ignore[union-attr]
        self.k = self.n - (Hx.shape[0] if Hx is not None else 0) - (Hz.shape[0] if Hz is not None else 0)
        self.Lx = CSSCode._compute_logical(self.Hz, self.Hx)
        self.Lz = CSSCode._compute_logical(self.Hx, self.Hz)

    def __hash__(self) -> int:
        """Compute a hash for the CSS code."""
        x_hash = int.from_bytes(self.Hx.tobytes(), sys.byteorder) if self.Hx is not None else 0
        z_hash = int.from_bytes(self.Hz.tobytes(), sys.byteorder) if self.Hz is not None else 0
        return hash(x_hash ^ z_hash)

    def __eq__(self, other: object) -> bool:
        """Check if two CSS codes are equal."""
        if not isinstance(other, CSSCode):
            return NotImplemented
        if self.Hx is None and other.Hx is None:
            assert self.Hz is not None
            assert other.Hz is not None
            return np.array_equal(self.Hz, other.Hz)
        if self.Hz is None and other.Hz is None:
            assert self.Hx is not None
            assert other.Hx is not None
            return np.array_equal(self.Hx, other.Hx)
        if (self.Hx is None and other.Hx is not None) or (self.Hx is not None and other.Hx is None):
            return False
        if (self.Hz is None and other.Hz is not None) or (self.Hz is not None and other.Hz is None):
            return False
        assert self.Hx is not None
        assert other.Hx is not None
        assert self.Hz is not None
        assert other.Hz is not None
        return bool(
            mod2.rank(self.Hx) == mod2.rank(np.vstack([self.Hx, other.Hx]))
            and mod2.rank(self.Hz) == mod2.rank(np.vstack([self.Hz, other.Hz]))
        )

    @staticmethod
    def _compute_logical(m1: npt.NDArray[np.int8] | None, m2: npt.NDArray[np.int8] | None) -> npt.NDArray[np.int8]:
        """Compute the logical matrix L."""
        if m1 is None:
            ker_m2 = mod2.nullspace(m2)  # compute the kernel basis of m2
            pivots = mod2.row_echelon(ker_m2)[-1]
            logs = np.zeros_like(ker_m2, dtype=np.int8)  # type: npt.NDArray[np.int8]
            for i, pivot in enumerate(pivots):
                logs[i, pivot] = 1
            return logs

        if m2 is None:
            return mod2.nullspace(m1).astype(np.int8)

        ker_m1 = mod2.nullspace(m1)  # compute the kernel basis of m1
        im_m2_transp = mod2.row_basis(m2)  # compute the image basis of m2
        log_stack = np.vstack([im_m2_transp, ker_m1])
        pivots = mod2.row_echelon(log_stack.T)[3]
        log_op_indices = [i for i in range(im_m2_transp.shape[0], log_stack.shape[0]) if i in pivots]
        return log_stack[log_op_indices]

    def get_x_syndrome(self, error: npt.NDArray[np.int8]) -> npt.NDArray[np.int8]:
        """Compute the x syndrome of the error."""
        if self.Hx is None:
            return np.empty((0, error.shape[0]), dtype=np.int8)
        return self.Hx @ error % 2

    def get_z_syndrome(self, error: npt.NDArray[np.int8]) -> npt.NDArray[np.int8]:
        """Compute the z syndrome of the error."""
        if self.Hz is None:
            return np.empty((0, error.shape[0]), dtype=np.int8)
        return self.Hz @ error % 2

    def check_if_logical_x_error(self, residual: npt.NDArray[np.int8]) -> bool:
        """Check if the residual is a logical error."""
        return bool((self.Lz @ residual % 2 == 1).any())

    def check_if_x_stabilizer(self, pauli: npt.NDArray[np.int8]) -> bool:
        """Check if the Pauli is a stabilizer."""
        assert self.Hx is not None
        return bool(mod2.rank(np.vstack((self.Hx, pauli))) == mod2.rank(self.Hx))

    def check_if_logical_z_error(self, residual: npt.NDArray[np.int8]) -> bool:
        """Check if the residual is a logical error."""
        return bool((self.Lx @ residual % 2 == 1).any())

    def check_if_z_stabilizer(self, pauli: npt.NDArray[np.int8]) -> bool:
        """Check if the Pauli is a stabilizer."""
        assert self.Hz is not None
        return bool(mod2.rank(np.vstack((self.Hz, pauli))) == mod2.rank(self.Hz))

    def stabilizer_eq_x_error(self, error_1: npt.NDArray[np.int8], error_2: npt.NDArray[np.int8]) -> bool:
        """Check if two X errors are in the same coset."""
        if self.Hx is None:
            return bool(np.array_equal(error_1, error_2))
        m1 = np.vstack([self.Hx, error_1])
        m2 = np.vstack([self.Hx, error_2])
        m3 = np.vstack([self.Hx, error_1, error_2])
        return bool(mod2.rank(m1) == mod2.rank(m2) == mod2.rank(m3))

    def stabilizer_eq_z_error(self, error_1: npt.NDArray[np.int8], error_2: npt.NDArray[np.int8]) -> bool:
        """Check if two Z errors are in the same coset."""
        if self.Hz is None:
            return bool(np.array_equal(error_1, error_2))
        m1 = np.vstack([self.Hz, error_1])
        m2 = np.vstack([self.Hz, error_2])
        m3 = np.vstack([self.Hz, error_1, error_2])
        return bool(mod2.rank(m1) == mod2.rank(m2) == mod2.rank(m3))

    def is_self_dual(self) -> bool:
        """Check if the code is self-dual."""
        if self.Hx is None or self.Hz is None:
            return False
        return bool(
            self.Hx.shape[0] == self.Hz.shape[0] and mod2.rank(self.Hx) == mod2.rank(np.vstack([self.Hx, self.Hz]))
        )

    def stabs_as_pauli_strings(self) -> tuple[list[str] | None, list[str] | None]:
        """Return the stabilizers as Pauli strings."""
        x_str = None if self.Hx is None else ["".join(["I" if x == 0 else "X" for x in row]) for row in self.Hx]
        z_str = None if self.Hz is None else ["".join(["I" if z == 0 else "Z" for z in row]) for row in self.Hz]
        return x_str, z_str

    def z_logicals_as_pauli_string(self) -> str:
        """Return the logical Z operator as a Pauli string."""
        return "".join(["I" if z == 0 else "Z" for z in self.Lx[0]])

    def x_logicals_as_pauli_string(self) -> str:
        """Return the logical X operator as a Pauli string."""
        return "".join(["I" if x == 0 else "X" for x in self.Lz[0]])

    @staticmethod
    def from_code_name(code_name: str, distance: int | None = None) -> CSSCode:
        r"""Return CSSCode object for a known code.

        The following codes are supported:
        - [[7, 1, 3]] Steane (\"Steane\")
        - [[15, 1, 3]] tetrahedral code (\"Tetrahedral\")
        - [[15, 7, 3]] Hamming code (\"Hamming\")
        - [[9, 1, 3]] Shore code (\"Shor\")
        - [[12, 2, 4]] Carbon Code (\"Carbon\")
        - [[9, 1, 3]] rotated surface code (\"Surface, 3\"), also default when no distance is given
        - [[25, 1, 5]] rotated surface code (\"Surface, 5\")
        - [[23, 1, 7]] golay code (\"Golay\")

        Args:
            code_name: The name of the code.
            distance: The distance of the code.
        """
        prefix = (Path(__file__) / "../").resolve()
        paths = {
            "steane": prefix / "steane/",
            "tetrahedral": prefix / "tetrahedral/",
            "hamming": prefix / "hamming/",
            "shor": prefix / "shor/",
            "surface_3": prefix / "rotated_surface_d3/",
            "surface_5": prefix / "rotated_surface_d5/",
            "golay": prefix / "golay/",
            "carbon": prefix / "carbon/",
        }

        distances = {
            "steane": (3, 3),
            "tetrahedral": (7, 3),
            "hamming": (3, 3),
            "shor": (3, 3),
            "golay": (7, 7),
            "surface_3": (3, 3),
            "surface_5": (5, 5),
            "carbon": (4, 4),
        }  # X, Z distances

        code_name = code_name.lower()
        if code_name == "surface":
            if distance is None:
                distance = 3
            code_name += f"_{distance}"

        if code_name in paths:
            hx = np.load(paths[code_name] / "hx.npy")
            hz = np.load(paths[code_name] / "hz.npy")

            if code_name in distances:
                x_distance, z_distance = distances[code_name]
                distance = min(x_distance, z_distance)
                return CSSCode(distance, hx, hz, x_distance=x_distance, z_distance=z_distance)

            if distance is None:
                msg = f"Distance is not specified for {code_name}"
                raise InvalidCSSCodeError(msg)
        msg = f"Unknown code name: {code_name}"
        raise InvalidCSSCodeError(msg)


class InvalidCSSCodeError(ValueError):
    """Raised when the CSS code is invalid."""



"""Estimate logical error rate for CSS state preparation circuits for a given code and physical error rate."""

from __future__ import annotations

import argparse
from pathlib import Path

from qiskit import QuantumCircuit

from mqt.qecc import CSSCode
from mqt.qecc.codes import HexagonalColorCode, SquareOctagonColorCode
from mqt.qecc.ft_stateprep import (
    NoisyNDFTStatePrepSimulator,
    gate_optimal_prep_circuit,
    gate_optimal_verification_circuit,
    heuristic_prep_circuit,
    heuristic_verification_circuit,
    naive_verification_circuit,
)


def main() -> None:
    """Run the logical error rate estimation for a given code and physical error rate."""
    available_codes = ["steane", "tetrahedral", "shor", "surface", "cc_4_8_8", "cc_6_6_6", "hamming", "carbon"]
    parser = argparse.ArgumentParser(description="Estimate logical error rate for CSS state preparation circuits")
    parser.add_argument(
        "code",
        type=str,
        help="Code for which to estimate logical error rate. Available codes: " + ", ".join(available_codes),
    )
    parser.add_argument("-p", "--p_error", type=float, help="Physical error rate")
    parser.add_argument("--zero_state", default=True, action="store_true", help="Synthesize logical |0> state.")
    parser.add_argument(
        "--plus_state", default=False, dest="zero_state", action="store_false", help="Synthesize logical |+> state."
    )
    parser.add_argument("-n", "--n_errors", type=int, default=500, help="Number of errors to sample")
    parser.add_argument("--exact_circ", default=False, action="store_true", help="Use exact synthesis")
    parser.add_argument("--heuristic_circ", dest="exact_circ", action="store_false", help="Use heuristic synthesis")
    parser.add_argument("--exact_ver", default=True, action="store_true", help="Use exact verification")
    parser.add_argument("--heuristic_ver", dest="exact_ver", action="store_false", help="Use heuristic verification")
    parser.add_argument("--naive_ver", default=False, action="store_true", help="Use naive verification")
    parser.add_argument("--no_ver", action="store_true", help="Use no verification")
    parser.add_argument(
        "-d", "--distance", type=int, default=3, help="Code Distance (only required for surface and color codes)"
    )
    parser.add_argument("--no_parallel_gates", default=False, action="store_true")

    args = parser.parse_args()
    code_name = args.code
    if "surface" in code_name:
        d = args.distance
        code = CSSCode.from_code_name("surface", d)
        code_name = f"rotated_surface_d{d}"
    elif "cc_4_8_8" in code_name:
        d = 5
        code = SquareOctagonColorCode(d)
    elif "cc_6_6_6" in code_name:
        d = 5
        code = HexagonalColorCode(d)
    elif code_name in available_codes:
        code = CSSCode.from_code_name(code_name)
    else:
        raise ValueError("Code " + code_name + " not available. Available codes: " + ", ".join(available_codes))

    prefix = (Path(__file__) / "../circuits/").resolve()
    sp_circ_name = "opt" if args.exact_circ else "heuristic"
    ver_circ_name = ("opt" if args.exact_ver else "heuristic") if not args.naive_ver else "naive"

    state_name = "zero" if args.zero_state else "plus"
    ft_name = "non_ft" if args.no_ver else "ft"
    circ_file = f"{state_name}_{ft_name}_{sp_circ_name}_{ver_circ_name}.qasm"

    # check if file exists
    if not (prefix / code_name / circ_file).exists():
        # create circuit
        circ = None
        if args.exact_circ:
            circ = gate_optimal_prep_circuit(code, zero_state=args.zero_state, max_timeout=600)
        else:
            circ = heuristic_prep_circuit(code, zero_state=args.zero_state)

        assert circ is not None
        if args.naive_ver:
            qc = naive_verification_circuit(circ)
        elif args.exact_ver:
            qc = gate_optimal_verification_circuit(circ, max_timeout=600)
        else:
            qc = heuristic_verification_circuit(circ)
    else:
        # load circuit from file
        qc = QuantumCircuit.from_qasm_file(prefix / code_name / circ_file)

    sim = NoisyNDFTStatePrepSimulator(
        qc, code=code, p=args.p_error, zero_state=args.zero_state, parallel_gates=not args.no_parallel_gates
    )
    res = sim.logical_error_rate(min_errors=args.n_errors)
    print(",".join([str(x) for x in res]))  # noqa: T201


if __name__ == "__main__":
    main()



"""Sparse code constructor for 3D and 4D HGP codes."""

from __future__ import annotations

import json
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import scipy.io as sio
from bposd.hgp import hgp
from ldpc import mod2
from scipy import sparse
from scipy.sparse import coo_matrix, csr_matrix

from . import code_constructor

if TYPE_CHECKING:
    from numpy.typing import NDArray


def is_all_zeros(array: NDArray[np.int_]) -> bool:
    """Check if array is all zeros."""
    return not np.any(array)


def sparse_all_zeros(mat: csr_matrix) -> bool:
    """Check if sparse matrix is all zeros."""
    mat.data %= 2
    return bool(mat.sum() == 0)


def run_checks_scipy(d_1: csr_matrix, d_2: csr_matrix, d_3: csr_matrix, d_4: csr_matrix) -> None:
    """Run checks on the boundary maps."""
    if not (sparse_all_zeros(d_1 @ d_2) and sparse_all_zeros(d_2 @ d_3) and sparse_all_zeros(d_3 @ d_4)):
        msg = "Error generating 4D code, boundary maps do not square to zero"
        raise RuntimeError(msg)


def generate_4d_product_code(
    a_1: csr_matrix,
    a_2: csr_matrix,
    a_3: csr_matrix,
    p: csr_matrix,
    checks: bool = True,
) -> tuple[csr_matrix, csr_matrix, csr_matrix, csr_matrix]:
    """Generate 4D HGP code."""
    r, c = p.shape

    id_r = sparse.identity(r, dtype=int)
    id_c = sparse.identity(c, dtype=int)
    id_n0 = sparse.identity(a_1.shape[0], dtype=int)
    id_n1 = sparse.identity(a_2.shape[0], dtype=int)
    id_n2 = sparse.identity(a_3.shape[0], dtype=int)
    id_n3 = sparse.identity(a_3.shape[1], dtype=int)

    d_1 = sparse.hstack((sparse.kron(a_1, id_r), sparse.kron(id_n0, p)))

    x = sparse.hstack((sparse.kron(a_2, id_r), sparse.kron(id_n1, p)))
    y = sparse.kron(a_1, id_c)
    dims = (y.shape[0], x.shape[1] - y.shape[1])
    nmat = csr_matrix(np.zeros(dims))
    z = sparse.hstack((nmat, y))
    d_2 = sparse.vstack((x, z))

    x = sparse.hstack((sparse.kron(a_3, id_r), sparse.kron(id_n2, p)))
    y = sparse.kron(a_2, id_c)
    dims = (y.shape[0], x.shape[1] - y.shape[1])
    mat = csr_matrix(np.zeros(dims))
    z = sparse.hstack([mat, y])
    d_3 = sparse.vstack((x, z))

    d_4 = sparse.vstack((sparse.kron(id_n3, p), sparse.kron(a_3, id_c)))

    if checks:
        run_checks_scipy(d_1, d_2, d_3, d_4)

    return d_1, d_2, d_3, d_4


def generate_3d_product_code(
    a_1: csr_matrix, a_2: csr_matrix, p: csr_matrix
) -> tuple[csr_matrix, csr_matrix, csr_matrix]:
    """Generate 3D HGP code."""
    r, c = p.shape

    id_r = sparse.identity(r, dtype=int)
    id_c = sparse.identity(c, dtype=int)
    id_n0 = sparse.identity(a_1.shape[0], dtype=int)
    id_n1 = sparse.identity(a_2.shape[0], dtype=int)
    id_n2 = sparse.identity(a_2.shape[1], dtype=int)

    d_1 = sparse.hstack((sparse.kron(a_1, id_r), sparse.kron(id_n0, p)))

    x = sparse.hstack((sparse.kron(a_2, id_r), sparse.kron(id_n1, p)))
    y = sparse.kron(a_1, id_c)
    dims = (y.shape[0], x.shape[1] - y.shape[1])
    z = sparse.hstack((csr_matrix(np.zeros(dims), dtype=int), y))
    d_2 = sparse.vstack((x, z))

    d_3 = sparse.vstack((sparse.kron(id_n2, p), sparse.kron(a_2, id_c)))

    if not (sparse_all_zeros(d_1 @ d_2) and sparse_all_zeros(d_2 @ d_3)):
        msg = "Error generating 3D code, boundary maps do not square to zero"
        raise RuntimeError(msg)

    return d_1, d_2, d_3  # mx, hx, hzT # hx, hzT, mzT


def create_outpath(codename: str) -> str:
    """Create output path for code."""
    path = f"/codes/generated_codes/{codename}/"
    Path(path).mkdir(parents=True, exist_ok=True)
    return path


def save_code(
    hx: csr_matrix,
    hz: csr_matrix,
    mz: csr_matrix,
    codename: str,
    lx: csr_matrix = None,
    lz: csr_matrix = None,
) -> None:
    """Save code to file."""
    path = create_outpath(codename)

    matrices: list[csr_matrix] = [hx, hz, mz, lx, lz]
    names = ["hx", "hz", "mz", "lx", "lz"]
    for mat, name in zip(matrices, names):
        if mat is not None:
            path_str = path + name
            try:
                np.savetxt(path_str + ".txt", mat.todense(), fmt="%i")
            except Exception:
                np.savetxt(path_str + ".txt", mat, fmt="%i")
            sio.mmwrite(
                path_str + ".mtx",
                coo_matrix(mat),
                comment="Field: GF(2)",
                field="integer",
            )


def run_compute_distances(codename: str) -> None:
    """Run compute distances bash script."""
    path = "/codes/generated_codes/" + codename
    subprocess.run(["bash", "compute_distances_3D.sh", path], check=False)  # noqa: S603, S607


def _compute_distances(hx: NDArray[np.int32], hz: NDArray[np.int32], codename: str) -> None:
    run_compute_distances(codename)
    _, n = hx.shape
    code_k = n - mod2.rank(hx) - mod2.rank(hz)
    with Path(f"/codes/generated_codes/{codename}/info.txt").open() as f:
        code_dict: dict[str, Any] = dict(
            line[: line.rfind("#")].split(" = ") for line in f if not line.startswith("#") and line.strip()
        )

    code_dict["n"] = n
    code_dict["k"] = code_k
    code_dict["dX"] = int(code_dict["dX"])
    code_dict["dZ"] = int(code_dict["dZ"])

    with Path(f"/codes/generated_codes/{codename}/code_params.txt").open("w") as file:
        file.write(json.dumps(code_dict))


def _store_code_params(hx: csr_matrix, hz: csr_matrix, codename: str) -> None:
    """Store code parameters in file."""
    code_dict = {}
    hx, hz = hx.todense(), hz.todense()
    _m, n = hx.shape
    code_k = n - mod2.rank(hx) - mod2.rank(hz)
    code_dict["n"] = n
    code_dict["k"] = code_k
    with Path(f"/codes/generated_codes/{codename}/code_params.txt").open("w") as file:
        file.write(json.dumps(code_dict))


def create_code(
    constructor: str,
    seed_codes: list[csr_matrix],
    codename: str,
    compute_distance: bool = False,
    compute_logicals: bool = False,
) -> None:
    """Create code."""
    # Construct initial 2 dim code
    if constructor == "hgp":
        code = hgp(seed_codes[0], seed_codes[1])
    else:
        msg = f"No constructor specified or the specified constructor {constructor} not implemented."
        raise ValueError(msg)

    # Extend to 3D HGP
    a1 = sparse.csr_matrix(code.hx)
    a2 = sparse.csr_matrix(code.hz.T)
    res = generate_3d_product_code(a1, a2, sparse.csr_matrix(seed_codes[2]))
    hx, hz_t, mz_t = res

    hz = hz_t.transpose()
    mz = mz_t.transpose()
    if compute_logicals:
        lx, lz = code_constructor._compute_logicals(hx.todense(), hz.todense())  # noqa: SLF001
        save_code(hx=hx, hz=hz, mz=mz, codename=codename, lx=lx, lz=lz)

    if compute_distance:
        _compute_distances(hx.todense(), hz.todense(), codename)

    else:
        _store_code_params(hx.todense(), hz.todense(), codename)



"""Methods for synthesizing fault tolerant state preparation circuits."""

from __future__ import annotations

from .simulation import LutDecoder, NoisyNDFTStatePrepSimulator
from .state_prep import (
    StatePrepCircuit,
    depth_optimal_prep_circuit,
    gate_optimal_prep_circuit,
    gate_optimal_verification_circuit,
    gate_optimal_verification_stabilizers,
    heuristic_prep_circuit,
    heuristic_verification_circuit,
    heuristic_verification_stabilizers,
    naive_verification_circuit,
)

__all__ = [
    "LutDecoder",
    "NoisyNDFTStatePrepSimulator",
    "StatePrepCircuit",
    "depth_optimal_prep_circuit",
    "gate_optimal_prep_circuit",
    "gate_optimal_verification_circuit",
    "gate_optimal_verification_stabilizers",
    "heuristic_prep_circuit",
    "heuristic_verification_circuit",
    "heuristic_verification_stabilizers",
    "naive_verification_circuit",
]
